//
// LIFReader.java
//

/*
OME Bio-Formats package for reading and converting biological file formats.
Copyright (C) 2005-@year@ UW-Madison LOCI and Glencoe Software, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package loci.formats.in;

import java.io.IOException;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import loci.common.DataTools;
import loci.common.RandomAccessInputStream;
import loci.common.services.DependencyException;
import loci.common.services.ServiceException;
import loci.common.services.ServiceFactory;
import loci.common.xml.XMLTools;
import loci.formats.CoreMetadata;
import loci.formats.FormatException;
import loci.formats.FormatReader;
import loci.formats.FormatTools;
import loci.formats.ImageTools;
import loci.formats.MetadataTools;
import loci.formats.meta.IMetadata;
import loci.formats.meta.MetadataStore;
import loci.formats.services.OMEXMLService;

import ome.xml.model.Channel;
import ome.xml.model.OME;
import ome.xml.model.Pixels;

/**
 * LIFReader is the file format reader for Leica LIF files.
 *
 * <dl><dt><b>Source code:</b></dt>
 * <dd><a href="http://dev.loci.wisc.edu/trac/java/browser/trunk/components/bio-formats/src/loci/formats/in/LIFReader.java">Trac</a>,
 * <a href="http://dev.loci.wisc.edu/svn/java/trunk/components/bio-formats/src/loci/formats/in/LIFReader.java">SVN</a></dd></dl>
 *
 * @author Melissa Linkert melissa at glencoesoftware.com
 */
public class LIFReader extends FormatReader {

  // -- Constants --

  public static final byte LIF_MAGIC_BYTE = 0x70;
  public static final byte LIF_MEMORY_BYTE = 0x2a;

  private static final Hashtable<String, Integer> CHANNEL_PRIORITIES =
    createChannelPriorities();

  private static Hashtable<String, Integer> createChannelPriorities() {
    Hashtable<String, Integer> h = new Hashtable<String, Integer>();

    h.put("red", new Integer(0));
    h.put("green", new Integer(1));
    h.put("blue", new Integer(2));
    h.put("cyan", new Integer(3));
    h.put("magenta", new Integer(4));
    h.put("yellow", new Integer(5));
    h.put("black", new Integer(6));
    h.put("gray", new Integer(7));
    h.put("", new Integer(8));

    return h;
  }

  private static final byte[][][] BYTE_LUTS = createByteLUTs();

  private static byte[][][] createByteLUTs() {
    byte[][][] lut = new byte[9][3][256];
    for (int i=0; i<256; i++) {
      // red
      lut[0][0][i] = (byte) (i & 0xff);
      // green
      lut[1][1][i] = (byte) (i & 0xff);
      // blue
      lut[2][2][i] = (byte) (i & 0xff);

      // cyan
      lut[3][1][i] = (byte) (i & 0xff);
      lut[3][2][i] = (byte) (i & 0xff);

      // magenta
      lut[4][0][i] = (byte) (i & 0xff);
      lut[4][2][i] = (byte) (i & 0xff);

      // yellow
      lut[5][0][i] = (byte) (i & 0xff);
      lut[5][1][i] = (byte) (i & 0xff);

      // gray
      lut[6][0][i] = (byte) (i & 0xff);
      lut[6][1][i] = (byte) (i & 0xff);
      lut[6][2][i] = (byte) (i & 0xff);
      lut[7][0][i] = (byte) (i & 0xff);
      lut[7][1][i] = (byte) (i & 0xff);
      lut[7][2][i] = (byte) (i & 0xff);
      lut[8][0][i] = (byte) (i & 0xff);
      lut[8][1][i] = (byte) (i & 0xff);
      lut[8][2][i] = (byte) (i & 0xff);
    }
    return lut;
  }

  private static final short[][][] SHORT_LUTS = createShortLUTs();

  private static short[][][] createShortLUTs() {
    short[][][] lut = new short[9][3][65536];
    for (int i=0; i<65536; i++) {
      // red
      lut[0][0][i] = (short) (i & 0xffff);
      // green
      lut[1][1][i] = (short) (i & 0xffff);
      // blue
      lut[2][2][i] = (short) (i & 0xffff);

      // cyan
      lut[3][1][i] = (short) (i & 0xffff);
      lut[3][2][i] = (short) (i & 0xffff);

      // magenta
      lut[4][0][i] = (short) (i & 0xffff);
      lut[4][2][i] = (short) (i & 0xffff);

      // yellow
      lut[5][0][i] = (short) (i & 0xffff);
      lut[5][1][i] = (short) (i & 0xffff);

      // gray
      lut[6][0][i] = (short) (i & 0xffff);
      lut[6][1][i] = (short) (i & 0xffff);
      lut[6][2][i] = (short) (i & 0xffff);
      lut[7][0][i] = (short) (i & 0xffff);
      lut[7][1][i] = (short) (i & 0xffff);
      lut[7][2][i] = (short) (i & 0xffff);
      lut[8][0][i] = (short) (i & 0xffff);
      lut[8][1][i] = (short) (i & 0xffff);
      lut[8][2][i] = (short) (i & 0xffff);
    }
    return lut;
  }

  // -- Fields --

  /** Offsets to memory blocks, paired with their corresponding description. */
  private Vector<Long> offsets;

  private int[][] realChannel;
  private int lastChannel = 0;

  // -- Constructor --

  /** Constructs a new Leica LIF reader. */
  public LIFReader() {
    super("Leica Image File Format", "lif");
    domains = new String[] {FormatTools.LM_DOMAIN};
  }

  // -- IFormatReader API methods --

  /* @see loci.formats.IFormatReader#isThisType(RandomAccessInputStream) */
  public boolean isThisType(RandomAccessInputStream stream) throws IOException {
    final int blockLen = 1;
    if (!FormatTools.validStream(stream, blockLen, true)) return false;
    return stream.read() == LIF_MAGIC_BYTE;
  }

  /* @see loci.formats.IFormatReader#get8BitLookupTable() */
  public byte[][] get8BitLookupTable() {
    FormatTools.assertId(currentId, true, 1);
    if (getPixelType() != FormatTools.UINT8 || !isIndexed()) return null;
    return lastChannel < BYTE_LUTS.length ? BYTE_LUTS[lastChannel] : null;
  }

  /* @see loci.formats.IFormatReader#get16BitLookupTable() */
  public short[][] get16BitLookupTable() {
    FormatTools.assertId(currentId, true, 1);
    if (getPixelType() != FormatTools.UINT16 || !isIndexed()) return null;
    return lastChannel < SHORT_LUTS.length ? SHORT_LUTS[lastChannel] : null;
  }

  /**
   * @see loci.formats.IFormatReader#openBytes(int, byte[], int, int, int, int)
   */
  public byte[] openBytes(int no, byte[] buf, int x, int y, int w, int h)
    throws FormatException, IOException
  {
    FormatTools.checkPlaneParameters(this, no, buf.length, x, y, w, h);

    if (!isRGB()) {
      int[] pos = getZCTCoords(no);
      lastChannel = realChannel[series][pos[1]];
    }

    long offset = offsets.get(series).longValue();
    int bytes = FormatTools.getBytesPerPixel(getPixelType());
    int bpp = bytes * getRGBChannelCount();

    long planeSize = (long) getSizeX() * getSizeY() * bpp;

    long nextOffset = series + 1 < offsets.size() ?
      offsets.get(series + 1).longValue() : in.length();
    int bytesToSkip = (int) (nextOffset - offset - planeSize * getImageCount());
    bytesToSkip /= getSizeY();
    if ((getSizeX() % 4) == 0) bytesToSkip = 0;

    in.seek(offset + planeSize * no);
    in.skipBytes(bytesToSkip * getSizeY() * no);

    if (bytesToSkip == 0) {
      readPlane(in, x, y, w, h, buf);
    }
    else {
      in.skipBytes(y * (getSizeX() * bpp + bytesToSkip));
      for (int row=0; row<h; row++) {
        in.skipBytes(x * bpp);
        in.read(buf, row * w * bpp, w * bpp);
        in.skipBytes(bpp * (getSizeX() - w - x) + bytesToSkip);
      }
    }

    // color planes are stored in BGR order
    if (getRGBChannelCount() == 3) {
      ImageTools.bgrToRgb(buf, isInterleaved(), bytes, getRGBChannelCount());
    }

    return buf;
  }

  /* @see loci.formats.IFormatReader#close(boolean) */
  public void close(boolean fileOnly) throws IOException {
    super.close(fileOnly);
    if (!fileOnly) {
      offsets = null;
      realChannel = null;
      lastChannel = 0;
    }
  }

  // -- Internal FormatReader API methods --

  /* @see loci.formats.FormatReader#initFile(String) */
  protected void initFile(String id) throws FormatException, IOException {
    super.initFile(id);
    in = new RandomAccessInputStream(id);
    offsets = new Vector<Long>();

    in.order(true);

    // read the header

    LOGGER.info("Reading header");

    byte checkOne = in.readByte();
    in.skipBytes(2);
    byte checkTwo = in.readByte();
    if (checkOne != LIF_MAGIC_BYTE && checkTwo != LIF_MAGIC_BYTE) {
      throw new FormatException(id + " is not a valid Leica LIF file");
    }

    in.skipBytes(4);

    // read and parse the XML description

    if (in.read() != LIF_MEMORY_BYTE) {
      throw new FormatException("Invalid XML description");
    }

    // number of Unicode characters in the XML block
    int nc = in.readInt();
    String xml = DataTools.stripString(in.readString(nc * 2));

    LOGGER.info("Finding image offsets");

    while (in.getFilePointer() < in.length()) {
      LOGGER.debug("Looking for a block at {}; {} blocks read",
        in.getFilePointer(), offsets.size());
      int check = in.readInt();
      if (check != LIF_MAGIC_BYTE) {
        throw new FormatException("Invalid Memory Block: found magic bytes " +
          check + ", expected " + LIF_MAGIC_BYTE);
      }

      in.skipBytes(4);
      check = in.read();
      if (check != LIF_MEMORY_BYTE) {
        throw new FormatException("Invalid Memory Description: found magic " +
          "byte " + check + ", expected " + LIF_MEMORY_BYTE);
      }

      long blockLength = in.readInt();
      if (in.read() != LIF_MEMORY_BYTE) {
        in.seek(in.getFilePointer() - 5);
        blockLength = in.readLong();
        check = in.read();
        if (check != LIF_MEMORY_BYTE) {
          throw new FormatException("Invalid Memory Description: found magic " +
            "byte " + check + ", expected " + LIF_MEMORY_BYTE);
        }
      }

      int descrLength = in.readInt() * 2;

      if (blockLength > 0) {
        offsets.add(new Long(in.getFilePointer() + descrLength));
      }

      in.seek(in.getFilePointer() + descrLength + blockLength);
    }
    initMetadata(xml);
    xml = null;
  }

  // -- Helper methods --

  /** Parses a string of XML and puts the values in a Hashtable. */
  private void initMetadata(String xml) throws FormatException, IOException {
    IMetadata omexml = null;
    try {
      ServiceFactory factory = new ServiceFactory();
      OMEXMLService service = factory.getInstance(OMEXMLService.class);
      omexml = service.createOMEXMLMetadata();
    }
    catch (DependencyException exc) {
      throw new FormatException("Could not create OME-XML store.", exc);
    }
    catch (ServiceException exc) {
      throw new FormatException("Could not create OME-XML store.", exc);
    }
    MetadataStore store = makeFilterMetadata();
    MetadataLevel level = getMetadataOptions().getMetadataLevel();
    LeicaHandler handler = new LeicaHandler(omexml, level);

    // the XML blocks stored in a LIF file are invalid,
    // because they don't have a root node

    xml = "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?><LEICA>" + xml +
      "</LEICA>";

    xml = XMLTools.sanitizeXML(xml);
    try {
      XMLTools.parseXML(xml, handler);
    }
    catch (IOException e) { }

    metadata = handler.getGlobalMetadata();
    Vector<String> lutNames = handler.getLutNames();

    core = handler.getCoreMetadata().toArray(new CoreMetadata[0]);

    // set up mapping to rearrange channels
    // for instance, the green channel may be #0, and the red channel may be #1
    realChannel = new int[getSeriesCount()][];
    int nextLut = 0;
    for (int i=0; i<getSeriesCount(); i++) {
      realChannel[i] = new int[core[i].sizeC];

      for (int q=0; q<core[i].sizeC; q++) {
        String lut = lutNames.get(nextLut++).toLowerCase();
        if (!CHANNEL_PRIORITIES.containsKey(lut)) lut = "";
        realChannel[i][q] = CHANNEL_PRIORITIES.get(lut).intValue();
      }

      int[] sorted = new int[core[i].sizeC];
      Arrays.fill(sorted, -1);

      for (int q=0; q<sorted.length; q++) {
        int min = Integer.MAX_VALUE;
        int minIndex = -1;
        for (int n=0; n<core[i].sizeC; n++) {
          if (realChannel[i][n] < min && !DataTools.containsValue(sorted, n)) {
            min = realChannel[i][n];
            minIndex = n;
          }
        }

        sorted[q] = minIndex;
      }
    }

    // remove any Channels that do not have an ID

    OMEXMLService service = null;
    try {
      service = new ServiceFactory().getInstance(OMEXMLService.class);
      if (service.isOMEXMLRoot(omexml.getRoot())) {
        OME root = (OME) omexml.getRoot();
        for (int i=0; i<getSeriesCount(); i++) {
          setSeries(i);
          Pixels img = root.getImage(i).getPixels();
          List<Channel> channels = img.copyChannelList();

          for (int c=0; c<channels.size(); c++) {
            Channel channel = channels.get(c);
            if (channel.getID() == null || c >= getEffectiveSizeC()) {
              img.removeChannel(channel);
            }
          }
        }
        setSeries(0);
        omexml.setRoot(root);
        service.convertMetadata(omexml, store);
      }
    }
    catch (DependencyException e) {
      LOGGER.trace("Failed to remove channels", e);
    }

    MetadataTools.populatePixels(store, this, true, false);
  }

}
